/*

Ser que escrevo o A* ?

Luis

2015-08-30: v0
2018-06-08: Current version

Changes relative to version v0:
In version 0, the concept of labled or tagged arc was not taken in
consideration. A solution to a search problem was a list of pairs
Nodei-Nodej (the arcs were not mentioned)
In the current version, arcs may or may not be considered at will. If
arcs are considered, a solution to a search problem will be a list of
arcs Arcij. If arcs are
not important or are not labled in any way, the solution will be the
same as in version v0.

Arcs might be important in cases in which one needs to know the
sequence of state transition actions that, when applied to the initial
node, generate a goal node (as in planning type problems).

Arcs will not be important, when A-star is used to find a path between
two graph nodes (the arcs represent the same conceptual action: move
from the origin node to the destination node).

Arcs will be important or not as determined by the expansion operation
passed as input to the A-star algorithm. If the expansion operation
generates a set o successors containing arc information, arcs will be
taken into account by the algorithm. Otherwise, they will be neglected.


*/

:- dynamic parent/2, state_tr_act/2, state_tr_act/3.

% A-Star Problem Solving
% astarps is a simpler interface to be used when the algorithm is used
% for problem solving, that is, for cases in which it is necessary to
% define the actions that perform state transitions (instead of just
% traversing unlabled arcs in a graph)
%
% To use this higher level simppler interface of the A-star algorithm,
% the programer must do the following:
% 1] define and specify all state transition actions, using the
% predicate state_tr_act/2 (or state_tr_act/3). The cost of each action
% is also specified using the predicate state_tr_act/2 (or state_tr_act/3). The first two
% arguments of the state transition actions must be the initial state
% and the result.
%
% 2] define the predicate that computes the expected cost to go from a
% node to the goal (heuristic) and specify it through the predicate
% astarps_heuristic_predicate/1; and
%
% 3] define the predicate that tests if a certain state is a goal state, using the predicate
% astarps_goal_test_predicate/1. There might be more than one
% ALTERNATIVE goal states.
%
astarps(Init, Solution, SCost):-
	astarps_heuristic_predicate(HeuristicPred),
	astarps_goal_test_predicate(GoalPred),
	astar(Init, astarps_expansion, HeuristicPred, astarps_cost, GoalPred, Solution, SCost).


astarps_expansion(State, Successors):-
	findall(
	  SOL_ACT-DState,
	  (astarps_state_tr_act(TR_ACT, SOL_ACT, _), TR_ACT =.. [_, State, DState | _], TR_ACT),
	  Successors).


% This predicate is used to return a state transition action, the term
% representing the action as it should appear in the solution and its
% cost.
% If the solution action term is not specified, the returned solution
% action term is the same as the specified state transition action.
astarps_state_tr_act(ACT, ACT, Cost):-
	state_tr_act(ACT, Cost),
	\+ state_tr_act(ACT, _, Cost).
astarps_state_tr_act(TR_ACT, SOL_ACT, Cost):-
	state_tr_act(TR_ACT, SOL_ACT, Cost),
	\+ state_tr_act(TR_ACT, Cost).

astarps_cost(ACT, Cost):-
	astarps_state_tr_act(_, ACT, Cost).


/*
% Examples
% In addition to definining the action move/3 (or move_up/2, move_down/2, move_left/2
% and move_right/2), the predicate hcost/2 and the predicate g/1, it is necessary to
% provide the following specifications:
%
% (for the room traversing problem)
state_tr_act(move(_, _, _), 1).

% Or (for the 8 puzzle domain)
state_tr_act(move_up(_, _), 1).
state_tr_act(move_down(_, _), 1).
state_tr_act(move_left(_, _), 1).
state_tr_act(move_right(_, _), 1).


astarps_heuristic_predicate(hcost).

astarps_goal_test_predicate(g).
*/


% A-Star API (astar/7 predicate)
% Expansion is a prolog predicate that takes a state and returns the
% set of its successors. Each sucessor is an ArcNode structure. An
% ArcNode structure is either a Node or a pair of an Arc and a Node:
% Arc-Node
%
% Heuristic is a prolog predicates that implements the h function.
%
% Cost is a prolog predicate that implements the g function
%
% GoalTest is a prolog prediccate that checks whether a given state is a goal
% state. It succeds or fails.
%
% Solution, to be computed by the algorithm, is a list containing the
% path from the initial node to the goal node. The path is a sequence of
% edges (i.e. arcs) containing information about either the pairs of
% nodes connected by each arc or the pair of connected nodes plus the
% tag of the arc (most often a state transition action).
%
% SCost, to be computed by the algorithm, is the cost of the
% solution
%
% To use the general purpose astar/7 program, one must provide the
% prolog definitions of Expansion, Heuristic, Cost and GoalTest.
%
astar(Init, Expansion, Heuristic, Cost, GoalTest, Solution, SCost):-
	node_heuristic_evaluation(Init, Heuristic, InitEval),
	astar([(Init, 0, InitEval)],[], Expansion, Heuristic, Cost, GoalTest, Goal, SCost),
	reconstruct_solution(Init, Goal, Solution),
	clean_parent_links.



% The astar/8 procedure implements most of the A* algorithm
% astar(Open, Closed, Expansion, Heuristic, Cost, GoalTest, Goal, GCost)
%
% Open: list of the nodes that have not yet been expanded. Each element
% in Open is a triple (State, Cost, Predicted evaluation (H))
% Closed: list of the nodes already expanded
% Goal, to be computed by the algorithm: Goal node
% GCost, to be computed by the algorithm: Cost of the goal node
%
% It creates a set of facts parent(Node, Parent) that enable
% reconstruct_solution/3 to return the solution.
%
% If the Open list is emty, clean the parent links and fail. Otherwise,
% a-star away
%
astar([], _, _, _, _, _, _, _):-
	!,
	clean_parent_links,
	fail.
astar(Open, Closed, Expansion, Heuristic, Cost, GoalTest, Goal, GCost):-
	best_node(Open, BestNode, NewOpen),
	aux_astar(NewOpen, Closed, Expansion, Heuristic, Cost, GoalTest, BestNode, Goal, GCost).

% aux_astar(NewOpen, Closed, Expansion, Heuristic, Cost, GoalTest, BestNode, Goal, GCost)
% BestNode is a triple (BestSate, GCost, HPrediction)
%
% The selected BestNode is a goal node
aux_astar(_, _, _, _, _, GoalTest, (Best, G, H), Best, BCost):-
	GOALTEST =.. [GoalTest, Best],
	call(GOALTEST),
	BCost is G+H, % Should be G since the node is a goal
	!.
% The selected BestNode is not a goal node, proceed
% There are successors
aux_astar(Open, Closed, Expansion, Heuristic, Cost, GoalTest, (Best, G, H), Goal, GCost):-
	EXPANSION =.. [Expansion, Best, Successors],
	call(EXPANSION),
	!,
	evaluated_nodes(Best, G, Successors, Heuristic, Cost, EvaluatedSuccessors),
	process_successors(Best, Open, [(Best, G, H)|Closed], EvaluatedSuccessors, UOpen, UClosed),
	astar(UOpen, UClosed, Expansion, Heuristic, Cost, GoalTest, Goal, GCost).
% There are no successors. This will never happen once the expansion
% predicate doesn't fail; it returns the empty list, instead
aux_astar(Open, Closed, Expansion, Heuristic, Cost, GoalTest, EvaluatedBest, Goal, GCost):-
	astar(Open, [EvaluatedBest|Closed], Expansion, Heuristic, Cost, GoalTest, Goal, GCost).


% The procedure of all decisions...
% For each sucessor...
%
% process_successors(Parent, Open, Closed, EvaluatedSuccessors, UOpen, UClosed)
% Open and Closed are lists of triples (Node, G, H)
% EvaluatedSuccessors is a list of triples (ArcNode, G, H) in which
% ArcNode is an ArcNode structure, and G and H is refer to the node in
% ArcNode.
%
% There is no equal Node with a better cost in either Open or Closed
process_successors(Parent, Open, Closed, [(ArcNode, AccG, H)|Successors], [(Node, AccG, H)|UOpen], UClosed):-
	node(ArcNode, Node),
	\+( (member((Node, GO, HO), Open), GO+HO < AccG+H) ),
	\+( (member((Node, GC, HC), Closed), GC+HC < AccG+H) ),
	!,
	update_parent_link(ArcNode, Parent),
	process_successors(Parent, Open, Closed, Successors, UOpen, UClosed).
% There is a better Node in open but not in closed
process_successors(Parent, Open, Closed, [(ArcNode, AccG, H)|Successors], UOpen, [(Node, AccG, H)|UClosed]):-
	node(ArcNode, Node),
	\+( (member((Node, GC, HC), Closed), GC+HC < AccG+H) ),
	!,
	process_successors(Parent, Open, Closed, Successors, UOpen, UClosed).
% There is a better Node in open and another one in closed
process_successors(Parent, Open, Closed, [_|Successors], UOpen, UClosed):-
	process_successors(Parent, Open, Closed, Successors, UOpen, UClosed).
% No more successors
process_successors(_, Open, Closed, [], Open, Closed).



% parent link recording section
%
% update_parent_link(ArcNode, Parent):
% ArcNode is an ArcNode structure, which is either a Node or a pair of
% an arc and a node: Arc-Node
% Parent is a node
%
% update_parent_link creates / updates the fact parent(Node, ArcParent)
% in which Node is the node in ArcNode and ArcParent is an ArcNode
% structure consisting of the arc possibly contained in ArcNode and the
% node Parent
%
update_parent_link(ArcNode, Parent):-
	parent_arcnode(ArcNode, Parent, Node, ArcParent),
	aux_update_parent_link(Node, ArcParent).

aux_update_parent_link(Node, ArcParent):-
	retract(parent(Node, _)),
	!,
	assert(parent(Node, ArcParent)).
aux_update_parent_link(Node, ArcParent):-
	assert(parent(Node, ArcParent)).


% Given an ArcNode structure and a parent node Parent, returns the node
% in ArcNode (Node) and the parent ArcNode with the same arc potentially
% in ArcNode, if any
parent_arcnode(Arc-Node, Parent, Node, Arc-Parent):- !.
parent_arcnode(Node, Parent, Node, Parent).





% Node evaluation section
node_heuristic_evaluation(Node, Heuristic, Eval):-
	HEURISTIC =.. [Heuristic, Node, Eval],
	call(HEURISTIC).

% This is not ok because, for many problems, one needs to know the name
% of the operation
node_accumulated_cost(Parent, Arc, Node, Cost, ParentAccCost, AccCost):-
	% PNCOST: Parent - Node Cost
	parent_arc_node_cost(Cost, Parent, Arc, Node, X),
	AccCost is X + ParentAccCost.

parent_arc_node_cost(Cost, _, Act, _, X):-
	nonvar(Act),
	!,
	PNCOST =.. [Cost, Act, X],
	call(PNCOST).
parent_arc_node_cost(Cost, Parent, _, Node, X):-
	PNCOST =.. [Cost, Parent, Node, X],
	call(PNCOST).


evaluated_nodes(Parent, ParentAccCost, [ArcNode|Successors], Heuristic, Cost, [(ArcNode, G, H)|EvaldSuccessors]):-
	node(ArcNode, Node), % Returns the node of an ArcNode structure
	arc(ArcNode, Arc),   % Returns the arc of an ArcNode structure or an unnamed variable
	node_accumulated_cost(Parent, Arc, Node, Cost, ParentAccCost, G),
	node_heuristic_evaluation(Node, Heuristic, H),
	evaluated_nodes(Parent, ParentAccCost, Successors, Heuristic, Cost, EvaldSuccessors).
evaluated_nodes(_, _, [], _, _, []).


% node(ArcNode, Node): receives an ArcNode structure and returns its
% node. An ArcNode structure is either a node or a pair of an arc and a
% node: Arc-Node.
%
node(_-Node, Node):- !.
node(Node, Node).


% arc(ArcNode, Arc): receives an ArcNode structure and returns its
% arc, if it is specified or an annamed variable otherwise. An ArcNode
% structure is either a node or a pair of an arc and a node: Arc-Node.
%
arc(Arc-_, Arc):- !.
arc(_, _).


/*
% Very inefficient
best_node(Open, (Best, G, H), NewOpen):-
	select((Best, G, H), Open, NewOpen),
	\+( (member((_, GN, HN), Open), GN+HN < G+H) ).
*/
% Tentative more efficient definition
best_node([N1, N2|Open], BestNode, NewOpen):-
	best_node([N2|Open], TMPBest, TmpOpen),
	better_node_and_remaining_list(N1, [N2|Open], TMPBest, [N1|TmpOpen], BestNode, NewOpen),
	!.
best_node([Node], Node, []).


% better_node_and_remaining_open(N1, L1, N2, L2, BestNode, NewList)
% Receives an element N1 and the remaini ng list L2, and another element
% N2 and the remaining list L2. Returns the better element and the
% corresponding remaining list.
better_node_and_remaining_list(N1, L1, N2, _, N1, L1):-
	% N1 is better than N2
	better_than_or_equal(N1, N2),
	!.
better_node_and_remaining_list(_, _, N2, L2, N2, L2).


% better_than_or_equal((Node1, G1, H1), (Node2, G2, H2))
better_than_or_equal((_, G1, H1), (_, G2, H2)):-
	(G1+H1) =< (G2+H2).



% Reconstruct solution
%
% This is not ok because, for many problems, one needs to know the name
% of the operation
reconstruct_solution(Start, Goal, Solution):-
	reconstruct_solution(Start, Goal, [], Solution).

reconstruct_solution(Start, Goal, Acc, Solution):-
	parent(Goal, ArcParent),
	solution_element(Goal, ArcParent, Parent, SolutionElement),
	reconstruct_solution(Start, Parent, [SolutionElement|Acc], Solution).
reconstruct_solution(Start, Start, Acc, Acc).


% solution_element(Node2, ArcNode1, Node1, SolutionElement)
% Given the destinantion node Node2 and an ArcNode structure of
% the origin node Node1, computes and returns the corresponding element
% to be added to the solution and the node in ArcNode1. The
% solution element, in any of its forms, always represents the directed
% arc connecting Node1 to Node2.
%S
% If ArcNode1 is a pair Arc-Node1, the solution element is Arc. If
% ArcNode is just a a node (Node1), the solution element is Node1-Node2.
%
solution_element(_, Arc-Node1, Node1, Arc):- !.
solution_element(Node2, Node1, Node1, Node1-Node2).


% Remove all facts parent links
clean_parent_links :-
	retractall(parent(_, _)).


/*

% Test
%
test1(Solution, Cost) :-
	clean_parent_links,
	astar(a, successors1, h, g1, goal, Solution, Cost).

test2(Solution, Cost) :-
	clean_parent_links,
	astar(a, successors2, h, g2, goal, Solution, Cost).

% TEST 1 (without arc lables)
suc1(a, b, 5).
suc1(a, c, 1).
suc1(a, d, 3).
% suc1(a, a, 0.25). % Causes no problem

suc1(c, d, 1).
% suc1(c, a, 1). % Causes no problem

suc1(d, e, 6).

suc1(e, f, 1).
% suc1(e, c, 0.25). % Causes no problem

suc1(N1, N2):- suc1(N1, N2, _).

g1(N1, N2, Cost):- suc1(N1, N2, Cost).


successors1(Node, Successors):-
	bagof(S, suc1(Node, S), Successors).


% TEST 2 (with arc lables)
suc2(a, b, act(a, b), 5).
suc2(a, c, act(a, c), 1).
suc2(a, d, act(a, d), 3).
% suc2(a, a, act(a, a), 0.25). % Causes no problem

suc2(c, d, act(c, d), 1).
% suc2(c, a, act(c, a), 1). % Causes no problem

suc2(d, e, act(d, e), 6).

suc2(e, f, act(e, f), 1).
% suc2(e, c, act(e, c), 0.25). % Causes no problem

suc2(N1, N2, Act):- suc2(N1, N2, Act, _).

% g2(Act, Cost):- suc2(N1, N2, Act, Cost).
g2(N1, N2, Cost):- suc2(N1, N2, _, Cost).


successors2(Node, Successors):-
	bagof(Act-S, suc2(Node, S, Act), Successors).



h(a, 3).
h(b, 6).
h(c, 4).
h(d, 1).
h(e, 1).
h(f, 0).

goal(f).

*/







